# Exemplary answer to Exercise 307. The magic happens
# in realprimes, which is based on a sieve found
# somewhere online. This takes a few  seconds to run.
# Pretty close to as fast as using the highly tuned
# primerange function found in the sympy module.
def realprimes(n):
  """Input n > 5, returns a list of primes, 3 < p < n.
  """
  if n <= 5:
      raise ValueError(f"realprimes needs a number > 5, but got {n}.")
  n, correction = n - n % 6 + 6, 2 - (n % 6 > 1)
  sieve = [True] * (n // 3)
  for i in range(1, int(n ** 0.5) // 3 + 1):
      if sieve[i]:
          k = 3 * i + 1 | 1
          sieve[k * k // 3 :: 2 * k] = [False] * ((n // 6 - k * k // 6 - 1) // k + 1)
          sieve[k * (k - 2 * (i & 1) + 4) // 3 :: 2 * k] = [False] * ((n // 6 - k * (k - 2 * (i & 1) + 4) // 6 - 1) // k + 1)
  return [3 * i + 1 | 1 for i in range(1, n // 3 - correction) if sieve[i]]

def list_the_gaps():
  """Finds the first million primes, and returns
     a list of all the gaps between them.
  """
  list_of_gaps = []
  prev = 3
  gap = 0
  # found at https://www.answers.com/Q/What_is_the_one_millionth_prime_number
  millionth_prime = 15485863

  for prime in list(realprimes(millionth_prime + 1)):
      gap = prime - prev
      prev = prime
      list_of_gaps.append(gap)
  return list_of_gaps

def make_frequency_table(gap_list):
  """Finds the frequency of each member of gap_list.
  """
  keys = sorted(sorted(set(gap_list), key = lambda ele: gap_list.count(ele)))
  freq = {}
  for i in keys:
      freq[i] = gap_list.count(i)
  return freq

# print(make_frequency_table(list_the_gaps()))
# prints
# {2: 86027, 4: 85832, 6: 146518, 8: 62750, 10: 80399, 12: 97826, 14: 52826, 16: 37688, 18: 66465, 20: 33777, 22: 29583, 24: 41731, 26: 18928, 28: 20591, 30: 33960, 32: 10100, 34: 10536, 36: 16301, 38: 7196, 40: 8600, 42: 11657, 44: 4492, 46: 3822, 48: 6305, 50: 3449, 52: 2412, 54: 4054, 56: 1862, 58: 1753, 60: 3194, 62: 910, 64: 980, 66: 1664, 68: 625, 70: 906, 72: 771, 74: 392, 76: 319, 78: 612, 80: 279, 82: 197, 84: 429, 86: 122, 88: 118, 90: 276, 92: 76, 94: 73, 96: 119, 98: 52, 100: 73, 102: 63, 104: 35, 106: 27, 108: 51, 110: 26, 112: 18, 114: 34, 116: 14, 118: 8, 120: 21, 122: 6, 124: 7, 126: 17, 128: 4, 130: 2, 132: 10, 134: 3, 136: 4, 138: 6, 140: 4, 142: 2, 144: 1, 146: 1, 148: 2, 150: 1, 152: 1, 154: 3}

# Exemplary answer to Exercise 314
from math import prod

def gcdxlcm(aInCSF, bInCSF):
  primes = list(map(lambda x:x[0], aInCSF))
  powers = lambda x:x[1]
  add = lambda x,y:x+y
  exp = lambda x,y:x**y
  aCSFpowers = list(map(powers, aInCSF))
  bCSFpowers = list(map(powers, bInCSF))
  gcdInCSF = list(zip(primes, map(min, aCSFpowers, bCSFpowers)))
  gcdCSFpowers = list(map(powers, gcdInCSF))
  gcd2mul = list(map(exp, primes, gcdCSFpowers))
  gcdProduct = prod(gcd2mul)
  lcmInCSF = list(zip(primes, map(max, aCSFpowers, bCSFpowers)))
  lcmCSFpowers = list(map(powers, lcmInCSF))
  lcm2mul = list(map(exp, primes, lcmCSFpowers))
  lcmProduct = prod(lcm2mul)
  gcdxlcmInCSF = list(zip(primes, map(add, aCSFpowers, bCSFpowers)))
  gcdxlcmPowers = list(map(powers, gcdxlcmInCSF))
  a2mul = list(map(exp, primes, aCSFpowers))
  b2mul = list(map(exp, primes, bCSFpowers))
  a = prod(a2mul)
  b = prod(b2mul)
  ab = a * b
  gcdxlcm2mul = list(map(exp, primes, gcdxlcmPowers))
  check = ab == prod(gcdxlcm2mul)
  return f"a = {aInCSF} = prod({a2mul}) = {a};\n\
b = {bInCSF} = prod({b2mul}) = {b};\n\
gcd(a, b) = {gcdInCSF} = prod({gcd2mul}) = {gcdProduct};\n\
lcm(a, b) = {lcmInCSF} = prod({lcm2mul}) = {lcmProduct};\n\
gcd * lcm = {gcdxlcmInCSF} = prod({gcdxlcm2mul}) =\n\
a * b = {a} * {b} = {ab} ({check})."

csf4a = [(2, 3), (3, 2), (5, 1), (7, 0)]
csf4b = [(2, 2), (3, 3), (5, 0), (7, 1)]
print(gcdxlcm(csf4a, csf4b))

# Exemplary answer to Exercise 315
def abr(n, b):
    d, m = n // b, n % b
    if (d == 0):
        return [m]
    else:
        l = abr(d, b)
        l.append(m)
        return l

# Exemplary answer to Exercise 317
def reconstruct_number_from_abr(list_of_coefficients, base):
    if not list_of_coefficients:
        return 0
    else:
        return list_of_coefficients[0] + (base *
        reconstruct_number_from_abr(list_of_coefficients[1:], base))

def from27to10(alphabetic_message):
    l = list(map(lambda n: int(ord(n) - ord('@')), alphabetic_message))
    l.reverse()
    return reconstruct_number_from_abr(l, 27)

def to27from10(decimal_number):
    l = list(map(lambda n: chr(n + ord('@')), abr(decimal_number, 27)))
    return ''.join(l)

# Exemplary answer to Exercise 319
def axVby(a, x, b, y):
  """
  Let a = 33 and b = 28, and 143 is certainly greater than 1, and doesn't divide
  either a or b, so choose a to rinse and repeat on.

     n   =     d   (q)   +    r
   -----------------------------
    33   =   143   (0)   +   33

  33 is less than 143, so skip to using 33 as a new n with 28 as d:

     n   =    d   (q)   +   r
   ---------------------------
    33   =   28   (1)   +   5

  And 5, which is less than 33, equals 33(1) + 28(-1), and doesn't divide either
  33 or 28, so continue with 33 as n and 5 as d:

     n   =   d   (q)   +   r
   --------------------------
    33   =   5   (6)   +   3

    3                              =
    33 - 5(6)                      =
    33(1) + 5(-6)                  = (verify with axVby(33, 1, 5, -6) -> 3)
    33(1) + (33(1) + 28(-1))(-6)   =
    33(1) + 33(1)(-6) + 28(-1)(-6) =
    33(1 + (1)(-6)) + 28(-1)(-6)   =
    33(-5) + 28(6)                 = (verify axVby(33, -5, 28, 6) -> 3)
    3

  Now 3, which is less than 5, equals 33(-5) + 28(6), but does divide 33, so use
  28 as n and 3 as d:

     n   =   d   (q)   +   r
   --------------------------
    28   =   3   (9)   +   1

    1                              =
    28 - 3(9)                      =
    28(1) + 3(-9)                  =
    28(1) + (33(-5) + 28(6))(-9)   =
    28(1) + 33(-5)(-9) + 28(6)(-9) =
    28(1) + 28(6)(-9) + 33(-5)(-9) =
    28(1 + (6)(-9)) + 33(-5)(-9)   =
    28(-53) + 33(45)               = (verify axVby(28, -53, 33, 45) -> 1)
    1

  So we have found an x=45 and y=-53 such that 33x + 28y = 1. Are those the only
  possible values for x and y? Hardly! The pattern shown below repeats forever.
  So what are the values of x and y that are smallest in absolute value?
  """
  return a * x + b * y

print([axVby(33, (45 + 28 * -1), 28, (-53 + 33 * 1)),
       axVby(33, (45 + 28 * -2), 28, (-53 + 33 * 2)),
       axVby(33, (45 + 28 * -3), 28, (-53 + 33 * 3))])
